;==============================================================================
;    serial.asm
;==============================================================================

;------------------------------------------------------------------------------
;
; Server monitor firmware - serial setup and interrupt routines
; File Version 1 - 31 Oct 10
; hairymnstr@gmail.com
;
; Copyright (C) 2010  Nathan Dumont
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
;
;------------------------------------------------------------------------------

list p=18f4520
include <p18f4520.inc>

; -- Externals From main.asm --------------------------------------------------
    extern      boot_loader_entry

; -- Externals From screen.asm ------------------------------------------------
    extern      screen_set_row
    extern      screen_print
    extern      screen_clear

; -- Externals From clock.asm -------------------------------------------------
    extern      clock_secs
    extern      clock_mins
    extern      clock_hours
    extern      clock_days
    extern      clock_month
    extern      clock_year

  UDATA

RX_MODE RES     1
; -- RX MODE CONSTANTS --------------------------------------------------------
RX_ACT  EQU     0
RX_BOOT EQU     1

RX_COUNT        RES     1
RX_CHECKSUM     RES     1
  
  CODE
;== Serial Console Functions ==================================================

;******************************************************************************
;*                                                                            *
;* serial_init - set up the serial port for PC comms                          *
;*                                                                            *
;* Inputs:  NONE                                                              *
;* Outputs: NONE                                                              *
;* Called:                                                                    *
;* Changes: SPBRG, RCSTA, TXSTA, BAUDCON, TRISC, RCON, PIR1, PIE1, INTCON     *
;*                                                                            *
;******************************************************************************

serial_init
    movlw       0x01
    movwf       FSR0H           ; set FSR0 to BANK1, RX Buffer
    clrf        FSR0L
    clrf        RX_MODE         ; set RX_MODE to look for a command
    ; first make sure that the RX and TX are both set input, hardware changes
    ; this once the UART is enabled
    movlw       b'11000000'
    iorwf       TRISC,f

    ; next set the UART up
    movlw       0x10
    movwf       SPBRG
    movlw       0x04
    movwf       SPBRGH          ; 9600 baud high speed 16 bit mode
    movlw       b'00100100'     ; enable sending, 8-bit, high speed baud rate
    movwf       TXSTA
    movlw       b'00010000'     ; enable receiving 8-bit
    movwf       RCSTA
    movlw       b'00001000'     ; no-auto sensing - 16 bit baud rate
    movwf       BAUDCON
    bsf         RCSTA,SPEN      ; enable the serial port

    ; set up interrupt on receive
    bcf         RCON, IPEN
    bcf         PIR1, RCIF
    bsf         PIE1, RCIE
    bsf         INTCON, PEIE
    bsf         INTCON, GIE    ; enable global interrupts
    return

;******************************************************************************
;*                                                                            *
;* serial_send - send the contents of W via the UART                          *
;*                                                                            *
;* Inputs:  Byte to send in W                                                 *
;* Outputs: NONE                                                              *
;* Called:                                                                    *
;* Changes: TXREG                                                             *
;*                                                                            *
;******************************************************************************

serial_send
    ; only do this if boot has finished and we're talking to python
    btfss       RX_MODE, RX_BOOT
    return
    btfss       PIR1,TXIF     ;test transmit interrupt flag
    bra         serial_send     ;if clear, TXREG is full
    movwf       TXREG         ;if set, TXREG is empty, so copy data into it
    return

;******************************************************************************
;*                                                                            *
;* serial_rx_int - UART receive interrupt routine                             *
;*                                                                            *
;* Inputs:  None                                                              *
;* Outputs: None                                                              *
;* Called:                                                                    *
;* Changes: PIR1, w                                                           *
;*                                                                            *
;******************************************************************************

serial_rx_int_error
    bcf         RCSTA,CREN              ;clear the CREN bit to clear errors
    movlw       0x15                    ;prepare a NAK symbol
    call        serial_send             ;and send it
    bsf         RCSTA,CREN              ;re-enable receiving
    movf        RCREG,w                 ;have to read the RCREG to clear the interrupt
    return

serial_rx_int
    movlw       b'00000110'
    andwf       RCSTA,w                 ;test if either of the error bits are set
    bnz         serial_rx_int_error     ;if result is not zero an error occurred

    movlw       0xAA                    ; put the start of message symbol in w
    xorwf       RCREG, w                ; xor with received byte
    bz          serial_rx_int_reset     ; if it was a start byte reset the comms link
    xorlw       0xAA                    ; otherwise put it back the way it was!
    
    btfss       RX_MODE, RX_ACT         ; check that the link is active
    return                              ; dump the bit in void if not

    movwf       POSTINC0                ; store the byte in the buffer
    xorwf       RX_CHECKSUM,f           ; update the checksum
    decfsz      RX_COUNT,f
    return                              ; if not the last byte all done for now

    movlw       0x00
    xorwf       RX_CHECKSUM,w           ; just to set the zero flag
    bnz         serial_rx_int_error     ; checksum failed, send nak

    movlw       0x06                    ; prepare ack
    call        serial_send

    bcf         RX_MODE, RX_ACT         ; no longer actively receiving until new start byte

    clrf        FSR0L                   ; point to the start of the RX buffer

    movf        POSTINC0,w              ; get the row number from the message
    andlw       0xF                     ; only 7 real rows but we use some "specials" too
    btfsc       WREG, 3                 ; see if this is one of the "special" rows, above 7
    goto        serial_rx_int_special
    call        screen_set_row

    movlw       d'21'                   ; count the characters on the line
    movwf       RX_COUNT

serial_rx_int_print_loop
    movf        POSTINC0,w              ; get a character from the message
    call        screen_print            ; print it
    decfsz      RX_COUNT,f              ; decrement the counter
    bra         serial_rx_int_print_loop
    return

serial_rx_int_reset
    clrf        FSR0L

    clrf        RX_CHECKSUM
    bsf         RX_MODE, RX_ACT

    movlw       d'23'
    movwf       RX_COUNT
    return

serial_rx_int_startup
    movlw       LOW startup_msg
    movwf       TBLPTRL
    movlw       HIGH startup_msg
    movwf       TBLPTRH
    clrf        TBLPTRU

    clrf        FSR0L
    movlw       d'22'
    movwf       RX_COUNT

serial_rx_int_startup_loop
    tblrd*+
    movf        TABLAT,w
    xorwf       POSTINC0, w
    btfss       STATUS, Z
    return                              ; exit this loop if it's not a match
    decfsz      RX_COUNT,f
    bra         serial_rx_int_startup_loop

    ; the special startup message has been received, from now on we can talk
    ; back on the serial link
    bsf         RX_MODE, RX_BOOT
    call        screen_clear            ; clear the booting message
    return

serial_rx_int_special
    xorlw       0x08
    bz          serial_rx_int_time
    xorlw       0x08
    xorlw       0x09
    bz          serial_rx_int_bootload
    xorlw       0x09
    xorlw       0x0F
    bz          serial_rx_int_startup
    goto        serial_rx_int_error

serial_rx_int_time
    ; special packet to update the local clock registers
    ; not implemented yet
    movf        POSTINC0, w
    movwf       clock_secs
    movf        POSTINC0, w
    movwf       clock_mins
    movf        POSTINC0, w
    movwf       clock_hours
    movf        POSTINC0, w
    movwf       clock_days
    movf        POSTINC0, w
    movwf       clock_month
    movf        POSTINC0, w
    movwf       clock_year
    movlw       0x06
    call        serial_send
    return

serial_rx_int_bootload
    ; the host requested we go into firmware update mode.
    movlw       0x06
    call        serial_send
    goto        boot_loader_entry

; -- End of interrupt routines ------------------------------------------------

startup_msg
    data        0x550F, 0x5555, 0x5555, 0x5555
    data        0x5555, 0x5555, 0x5555, 0x5555
    data        0x5555, 0x5555, 0x5555

    GLOBAL      serial_rx_int
    GLOBAL      serial_init
    end
